/*
  GNUJSP - a free JSP1.0 implementation
  Copyright (C) 1999, Yaroslav Faybishenko <yaroslav@cs.berkeley.edu>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package gnu.jspengine.xml;

import java.io.IOException;
import java.io.StringWriter;
import java.io.BufferedWriter;
import java.net.URLEncoder;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;

import gnu.jspengine.Emitter;
import gnu.jspengine.JspMsg;
import gnu.jspengine.JspNode;
import gnu.jspengine.Pos;
import gnu.jspengine.ParseException;
import gnu.jspengine.JspConfig;

import gnu.jspengine.xml.DOMSink;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;

/**
 * create DOM tree representation of internal data structure
 */

public class DOMEmitter extends Emitter 
    implements JspMsg
{
    public static final String HTTP_JSP_PAGE_IMPL = "gnu.jspengine.HttpJspPageImpl";
    public static final String HTTP_JSP_PAGE_UTIL = "gnu.jspengine.HttpJspPageUtil";
    public static final String GNU_JSP_PAGE = "gnu.jspengine.GnuJspPage";

    public static final void emitPos(BufferedWriter w, JspNode node) 
	throws IOException
    {
	DOMSink ds = (DOMSink) w;
	Pos pos = node.getPos();

	if(pos != null) {
	    ds.appendChild(ds.getDocument().createComment(pos.toJavaCode()));
	}
    }
    private void setNewAttribute(DOMSink ds, Element el, String name, String value) {
	if(value != null) {
	    el.setAttributeNode(ds.getDocument().createAttribute(name));
	    el.setAttribute(name, value);
	}
    }

    public void emitBeginPage (BufferedWriter w, JspNode node) throws IOException {
	NamedNodeMap attrs = null;
	DOMSink ds = (DOMSink) w;

	ds.getRoot().appendChild(ds.getDocument().createComment("Generated by gnujsp DOMEmitter on "+new Date ()));

	Element el = ds.getDocument().createElement("jsp:directive.page");
	setNewAttribute(ds, el, "info", node.getAttribute("info"));
	setNewAttribute(ds, el, "import", node.getAttribute("import"));
	setNewAttribute(ds, el, "isThreadSafe", node.getAttribute("isThreadSafe"));
	setNewAttribute(ds, el, "extends", node.getAttribute("extends"));
	setNewAttribute(ds, el, "contentType", node.getAttribute("contentType"));
	setNewAttribute(ds, el, "errorPage", node.getAttribute("errorPage"));
	setNewAttribute(ds, el, "isErrorPage", node.getAttribute("isErrorPage"));
	setNewAttribute(ds, el, "session", node.getAttribute("session"));
	setNewAttribute(ds, el, "buffer", node.getAttribute("buffer"));
	setNewAttribute(ds, el, "autoFlush", node.getAttribute("autoFlush"));
	setNewAttribute(ds, el, "language", node.getAttribute("language"));
	attrs = el.getAttributes();
	if(attrs != null && attrs.getLength() > 0) {
	    ds.getRoot().appendChild(el);
	}
	String declaration = node.getAttribute("declaration");
	if(declaration != null) {
	    Element decl = ds.getDocument().createElement("jsp:declaration");
	    decl.appendChild(ds.getDocument().createCDATASection(declaration));
	    //	    el.appendChild(decl);
	    // FIXME: check if declaration is inside page directive
	    ds.getRoot().appendChild(decl);
	}
    }
   
    public void emitIncludeDirective (BufferedWriter w, JspNode n, 
				      String file) throws IOException {
	DOMSink ds = (DOMSink) w;

	Element el = ds.getDocument().createElement("jsp:directive.include");
	setNewAttribute(ds, el, "file", n.getAttribute("file"));
	ds.getRoot().appendChild(el);
    }

    public void emitBeginScriptlet (BufferedWriter w, JspNode n) throws IOException {
	DOMSink ds = (DOMSink) w;

	emitPos(w, n);
	Element el = ds.getDocument().createElement("jsp:scriptlet");
	ds.pushNode(el);
    }

    public void emitScriptlet (BufferedWriter w, JspNode n, String code) throws IOException {
	DOMSink ds = (DOMSink) w;

	// FIXME: proper escaping needed! (CDATA!)
	ds.appendChild(ds.getDocument().createCDATASection(code));
    }

    public void emitEndScriptlet (BufferedWriter w) throws IOException {
	DOMSink ds = (DOMSink) w;

	ds.popNode();
    }

    public void emitBeginPrintExprCall (BufferedWriter w, JspNode n) throws IOException {
	DOMSink ds = (DOMSink) w;

	emitPos(w, n);
	Element el = ds.getDocument().createElement("jsp:expression");
	ds.pushNode(el);
    }

    public void emitPrintExprCall (BufferedWriter w, JspNode n, String expr) throws IOException {
	DOMSink ds = (DOMSink) w;

	// FIXME: proper escaping needed! (CDATA!)
	ds.appendChild(ds.getDocument().createCDATASection(expr));
    }

    public void emitEndPrintExprCall (BufferedWriter w) throws IOException {
	DOMSink ds = (DOMSink) w;

	ds.popNode();
    }

    public void emitTemplateText (BufferedWriter w, JspNode n, String t) throws IOException {
	DOMSink ds = (DOMSink) w;

	emitPos(w, n);
	// FIXME: proper escaping needed! (CDATA!)
	ds.appendChild(ds.getDocument().createCDATASection(t));
    }

    public void emitForward (BufferedWriter w, JspNode node, String page) throws IOException {
	DOMSink ds = (DOMSink) w;

	emitPos(w, node);
	Element el = ds.getDocument().createElement("jsp:forward");
	setNewAttribute(ds, el, "page", node.getAttribute("page"));
	writeQueryString(ds, node, el);
	ds.appendChild(el);
    }

    public void emitInclude (BufferedWriter w, JspNode node, String page, String flush) throws IOException {
	DOMSink ds = (DOMSink) w;

	emitPos(w, node);
	Element el = ds.getDocument().createElement("jsp:include");
	setNewAttribute(ds, el, "page", node.getAttribute("page"));
	setNewAttribute(ds, el, "flush", node.getAttribute("flush"));
	writeQueryString(ds, node, el);
	ds.appendChild(el);
    }

    private void writeQueryString(DOMSink ds, JspNode node, Element el) throws IOException {
	// If any parameters were set, generate a query string.
	// values are request-time expressions.
	Hashtable params = node.getParameters();
	if (params != null) {
	    Document doc = ds.getDocument();
	    Element paramsEl = doc.createElement("jsp:params");
	    el.appendChild(paramsEl);

	    Enumeration e = params.keys();
	    String name;
	    String[] values;
	    boolean first = true;

	    while (e.hasMoreElements()) {
		name = (String) e.nextElement();
		values = (String[]) params.get(name);
		for (int i = 0; i < values.length; i++) {
		    // FIXME: quoting!
		    Element param = doc.createElement("jsp:param");
		    setNewAttribute(ds, param, name, values[i]);
		    paramsEl.appendChild(param);
		} // for each value
	    } // for each key
	} // if non-null
    }
	
    public void emitBeginUseBean (BufferedWriter w, JspNode node, String id, String scope,
			   String type, String clas, String beanName)
	throws IOException
    {
	DOMSink ds = (DOMSink) w;

	emitPos(w, node);
	Element el = ds.getDocument().createElement("jsp:usebean");
	ds.pushNode(el);
	setNewAttribute(ds, el, "type", type);
	setNewAttribute(ds, el, "class", clas);
	setNewAttribute(ds, el, "id", id);
	setNewAttribute(ds, el, "scope", scope);
	setNewAttribute(ds, el, "beanName", beanName);
    }

    public void emitEndUseBean (BufferedWriter w) throws IOException {
	DOMSink ds = (DOMSink) w;

	ds.popNode();
    }

    public void emitGetProperty (BufferedWriter w, JspNode node, String name, 
				 String property) 
	throws IOException
    {
	DOMSink ds = (DOMSink) w;

	emitPos(w, node);
	Element el = ds.getDocument().createElement("jsp:getProperty");
	ds.appendChild(el);
	setNewAttribute(ds, el, "name", name);
	setNewAttribute(ds, el, "property", property);
    }

    public void emitSetProperty (BufferedWriter w, JspNode node,
				 String name, String property, 
				 String param, String value)
	throws IOException
    {
	DOMSink ds = (DOMSink) w;

	// FIXME: quoting of runtime args!

	emitPos(w, node);
	Element el = ds.getDocument().createElement("jsp:setProperty");
	ds.appendChild(el);
	setNewAttribute(ds, el, "name", name);
	setNewAttribute(ds, el, "property", property);
	setNewAttribute(ds, el, "param", param);
	setNewAttribute(ds, el, "value", value);
    }

    public void emitEndPage (BufferedWriter w, JspNode n) throws IOException {
	DOMSink ds = (DOMSink) w;
	ds.writeDOM();
    }

    /**
     * Converts a string into a format that can be used in a 
     * print() function, surrounding it with double quotes and
     * escaping any necessary control characters.
     */
    void enquote(BufferedWriter w, String t) throws IOException {
	w.write('"');
	char c;
	for (int i = 0; i < t.length(); i++) {
	    switch (c = t.charAt(i)) {
	    case '\t': 
		w.write ("\\t");
		break;
	    case '\n': 
		w.write ("\\n");
		break;
	    case '\r': 
		w.write ("\\r"); 
		break;
	    case '\\': 
		w.write ("\\\\"); 
		break;
	    case '\"': 
		w.write ("\\");
	    default: 
		w.write (c);
	    }
	}
	w.write('"');
    }

    /**
     * Translates a String argument, encoding any necessary request-time
     * arguments, and prepending and appending a double-quote character.
     * Also, escape any double-quotes, etc., in the string.
     *
     * Example 1:               "foo" --> "\"foo\""
     * Example 2:       "<%= name %>" --> "\"" + name + "\"" 
     *
     * Output is suitable as the parameter to out.print(_____)
     */
    String inlineAttribute (Pos pos, String val) throws IOException {
	int pos1 = val.indexOf("<%=");
	int pos2 = val.indexOf("%>");
	if ((pos1 == 0) && (pos2 == val.length() - 2)) {
	    return val.substring(3, val.length() - 2);
	}
	if ((pos1 >= 0) && (pos1 < pos2)) {
	    throw new ParseException
		(pos, 
		 JspConfig.getLocalizedMsg
		 (ERR_sp10_2_12_1_invalid_value_expression_and_string)
		 +": " + val );
	}

	StringBuffer b = new StringBuffer();
	int prev = 0;
	int i = 0;
	do {
	    pos1 = val.indexOf("<\\%", i);
	    pos2 = val.indexOf("%\\>", i);
	    if (pos1 == pos2) break; // both are -1, we're done.
	    i = 1 + (((pos2 == -1) || ((pos1 >= 0) && (pos1 < pos2))) 
		     ? pos1 : pos2);
	    b.append(val.substring(prev, i));
	    prev = ++i;
	} while (true);
	b.append(val.substring(prev));
	StringWriter w = new StringWriter();
	BufferedWriter bw = new BufferedWriter(w);
	enquote(bw,b.toString());
	bw.flush();
	return w.toString();
    }
}
